#include "component.h"

#include "collection.h"

constexpr std::string_view masonJSON("mason.json");
constexpr std::string_view dependencies("dependencies");
constexpr std::array<std::string_view, Component::done + 1> stringStates {
    "initial",
    "reading",
    "ready",
    "building",
    "error",
    "done"
};

constexpr std::array<std::string_view, Component::unknown + 1> stringTypes {
    "file",
    "directory",
    "mason",
    "unknown"
};

Component::Component(
    const std::filesystem::path& path,
    const std::weak_ptr<Collection>& collection,
    const std::shared_ptr<Logger>& logger
):
    Loggable(logger),
    state(initial),
    type(unknown),
    collection(collection),
    location(path),
    manifest(std::nullopt)
{}

void Component::read() {
    if (state != initial)
        throw WrongState(state, "read");

    state = reading;

    std::filesystem::file_status status = std::filesystem::status(location);
    switch (status.type()) {
        case std::filesystem::file_type::directory:
            type = directory;
            state = ready;
            tryReadingBuildScenarios();
            break;
        case std::filesystem::file_type::regular:
            type = file;
            state = ready;
            break;
        default:
            warn(location.string() + " doesn't appear to be a file nor a directory");
            state = error;
            break;
    }
}

void Component::tryReadingBuildScenarios() {
    if (readAsMason())
        return;
}

bool Component::readAsMason() {
    std::filesystem::path masonManifest = location/masonJSON;
    try {
        std::ifstream file(masonManifest);
        if (file.is_open())
            manifest = nlohmann::json::parse(file);
        else
            minor("Couldn't open " + masonManifest.string());
    } catch (const nlohmann::json::exception& e) {
        major("Couldn't parse " + masonManifest.string());
    }

    if (manifest.has_value()) {
        const nlohmann::json& mnfst = manifest.value();
        if (!mnfst.is_object())
            return errorScenario(masonManifest.string() + " is unexpected root type");

        nlohmann::json::const_iterator itr = mnfst.find(dependencies);
        if (itr == mnfst.end()) {
            info(masonManifest.string() + " doesn't have dependencies");
        } else {
            const nlohmann::json& deps = *itr;
            if (!readMasonDependencies(deps, masonManifest))
                return false;
        }

        type = mason;
        info(location.string() + " seems to be a valid mason project");
        return true;
    }

    return false;
}

bool Component::readMasonDependencies(const nlohmann::json& deps, const std::string& manifestPath) {
    if (!deps.is_array())
        return errorScenario(manifestPath + " dependencies are not organized as an array");

    for (const nlohmann::json& dep : deps) {
        std::shared_ptr<Collection> col = collection.lock();
        switch (dep.type()) {
            case nlohmann::json::value_t::string:
                col->addSource(dep);
                break;
            case nlohmann::json::value_t::object: {
                nlohmann::json::const_iterator ditr = dep.find("path");
                if (ditr == dep.end())
                    return errorScenario(manifestPath + " object of unexpected format in dependecies");

                nlohmann::json path = *ditr;
                if (!path.is_string())
                    return errorScenario(manifestPath + " object of unexpected format in dependecies");

                col->addSource(path);
            }
                break;
            default:
                return errorScenario(manifestPath + " has unexpected entries in dependecies");
        }
    }

    return true;
}


bool Component::errorScenario(const std::string& message) {
    major(message);
    manifest = std::nullopt;
    return false;
    return false;
}

void Component::build(const std::filesystem::path& destination, const std::string& target) {
    if (state != ready)
        throw WrongState(state, "build");

    state = building;
    info("Building " + location.string() + " to " + destination.string());

    switch (type) {
        case file:
            buildAsFile(destination);
            break;
        case directory:
            buildAsDiractory(destination);
            break;
        case mason:
            buildAsMason(destination, target);
            break;
        default:
            break;
    }

    state = done;
}

void Component::buildAsFile(const std::filesystem::path& destination) {
    std::filesystem::copy(location, destination/location.filename());
}

void Component::buildAsDiractory(const std::filesystem::path& destination) {
    std::filesystem::copy(location, destination/location.filename(),
                          std::filesystem::copy_options::recursive |
                          std::filesystem::copy_options::overwrite_existing
    );
}

std::string Component::getLibrariesPath() const {
    //TODO may be it makes sence to cache
    if (type != mason)
        throw WrongType(type, "getLibrariesPath");

    const nlohmann::json& mnfst = manifest.value();
    nlohmann::json::const_iterator itr = mnfst.find("libraries");
    if (itr == mnfst.end())
        return "";

    const nlohmann::json& libs = *itr;
    if (libs.is_string())
        return libs;

    warn("Mason manifest " + (location/masonJSON).string()
        + " has field \"libraries\" which is not string, ignoring it");
    return "";
}

std::filesystem::path Component::getLocation() const {
    return location;
}

bool Component::successfullyRead() const {
    return state > reading && state != error;
}

void Component::buildAsMason(const std::filesystem::path& destination, const std::string& target) {
    warn("mason build is not implemented yet");
}

Component::State Component::getState() const {
    return state;
}

Component::Type Component::getType() const {
    return type;
}

Component::WrongState::WrongState(State state, const std::string& action):
    std::runtime_error("An attempt to perform action \"" + action
                        + "\" on a wrong component state \"" + stringStates[state].data() + '\"')
{}

Component::WrongType::WrongType(Type type, const std::string& action):
    std::runtime_error("An attempt to perform an action \"" + action
                        + "\" on a wrong component type \"" + stringTypes[type].data() + '\"')
{}
